/*
** This program is part of the FLoP project
**
** Copyright (C) 2002-2006 Dirk Geschke <Dirk@geschke-online.de>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/*************************************************************************/
/*                                                                       */
/* Program to receive network data from a snort database based on the    */
/* sensor id (sid) and counter id (cid). From this data a network packet */
/* in pcap format is written to a file or stdout (filename = '-'). This  */
/* file can be used to analyze it via tcpdump or ethereal to use their   */
/* protocol analyzing features. Simply to view the ascii payload may not */
/* be enough.                                                            */
/*                                                                       */
/* Note: You need an extended database scheme in contrast to the normal  */
/*       snort database used in conjunction with acid                    */
/*                                                                       */
/*************************************************************************/

static char *rcsid="$Id: getpacket.c,v 1.36 2006/05/09 19:34:09 geschke Exp $";

#include "FLoPconfig.h"
#undef PACKAGE
#undef VERSION
#undef PACKAGE
#undef VERSION
#undef PACKAGE_BUGREPORT
#undef PACKAGE_NAME
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME
#undef PACKAGE_VERSION

#ifndef HAVE_U_INT8_T
#define u_int8_t  uint8_t
#endif
#ifndef HAVE_U_INT16_T
#define u_int16_t uint16_t
#endif
#ifndef HAVE_U_INT32_T
#define u_int32_t uint32_t
#endif

#include <sys/types.h>
#include <stdio.h>
#include <sys/un.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_PCAP_HEADER
#include <pcap.h>
#endif

#include "snort.h"
#include "plugbase.h"

#define STRLEN 256
#define MAX_QUERY_LENGTH 8192
#define MAXBUFFSIZE 65536 /* at least one complete TCP session */
#define DG_LITTLE_ENDIAN 0
#define DG_BIG_ENDIAN 1

#ifdef DEBUG
#define DEBUG_AP
#define DEBUG_CMDLINE
#define DEBUG_DB
#define DEBUG_GP
#define DEBUG_READCONFIG
#endif
#ifndef UNIX_PATH_MAX
#define UNIX_PATH_MAX   108
#endif

#ifdef MYSQL_DB
#include <mysql.h>
#endif

#ifdef POSTGRES
#include <libpq-fe.h>
#endif

#define DB_MYSQL 1
#define DB_POSTGRES 2

typedef struct _DG_pcap_timeval
{
	u_int32_t tv_sec;
	u_int32_t tv_usec;
} DG_pcap_timeval;

typedef struct _DG_pcap_pkthdr
{
        DG_pcap_timeval ts;
	u_int32_t caplen;
	u_int32_t len;
} DG_pcap_pkthdr;

typedef struct _GPoptions
{
  int32_t sid;
  int32_t cid;
  int32_t nextcid;
  int32_t reference;
  char ConfigFile[STRLEN];
  char DumpFile[STRLEN];
  char DBuser[STRLEN];
  char DBname[STRLEN];
  int DBtype;
  int DBencoding;
  int followtags;
  int alltagged;
  int disabletags;
  char DBpassword[STRLEN];
  char SocketName[STRLEN];
#ifdef MYSQL_DB
    MYSQL * m_sock;
    MYSQL_RES * m_result;
    MYSQL_ROW m_row;
#endif
#ifdef POSTGRES
   PGconn     *p_sock;
   PGresult   *p_result;
#endif
} GPoptions;


/**** GLOBAL VALUES ***/

static char *select_ptr=NULL; /* Memory is allocted in DatabaseInit! */
static char *insert_ptr=NULL; /* Memory is allocted in DatabaseInit! */
static char *dbuffer=NULL;    /* Needed in Database() */
static char *tbuffer=NULL;  /* Needed in Database() to store palyoad */
static char *pbuffer=NULL;  /* Needed in Database() to store palyoad */
static char VendorFile[STRLEN]="vendor.map";
static char link_layer_prot[STRLEN]="DLT_EN10MB";


/* Needed across BuildPacketHeader(), AppendPacket() and ClosePacket() */
static FILE *dumpfd=NULL; 
#ifdef HAVE_PCAP_HEADER
static pcap_t *pd=NULL;
static pcap_dumper_t *p=NULL;
#endif

static FILE *output;

static int Verbose=0;
static int Count=0;
static int Info=0;
static int VendorOpt=0;
char endian;

/***** PROTOTYPES *****/

void CmdLine(int argc, char* argv[], GPoptions *opt);
void Usage(char *argv);
int Select(char * query, GPoptions * data);
char * Select1(char * query, GPoptions * data);
int Insert(char * query, GPoptions * data);
int Connect(GPoptions * data);
void Disconnect(GPoptions * data);
int DatabaseInit(GPoptions * data);
void ReadConfig(GPoptions* Options);
void ArgCheck(char *arg);
int GetPacket(GPoptions *Options);
int BuildPacketHeader(GPoptions* Options);
int AppendPacket(GPoptions* Options);
int ClosePacket(GPoptions* Options);
unsigned int hextoint(char *t, int len);
char * GetVendor(char *t);
char CheckEndian();
int base64decode(unsigned char *line);
int base64tohex(char *buffer);
int bin2hex(unsigned char *data,int length);
int base2hex(char *buffer);


/****** MAIN *****/

int 
main(int argc, char* argv[])
{
  GPoptions Options;
  int res;

  output=stdout;
  bzero(&Options,sizeof(Options));

  endian=CheckEndian();

#ifdef DEBUG
  if (endian == DG_LITTLE_ENDIAN)
    {
      fprintf(output,"Little-Endian System\n");
    }
  else
    {
      fprintf(output,"Big-Endian System\n");
    }
#endif 

  /* preset some values */
  snprintf(Options.DumpFile,STRLEN,"/var/tmp/dump");
  snprintf(Options.ConfigFile,STRLEN,"%s/getpacket.conf",PATH_CONFIG);

  /* get the command line arguments */

  snprintf(VendorFile,STRLEN,"%s/vendor.map",PATH_CONFIG);

  CmdLine(argc,argv,&Options);

#ifdef DEBUG
  fprintf(output,"sid: %i, cid: %i\n",Options.sid,Options.cid);
#endif

  /* Check minimal necessary values */
  if (Options.sid < 1)
    {
      fprintf(stderr,"You have to set SID to a positive value: %i\n",
	      Options.sid);
      fprintf(stderr,"Use option -S to set the SID.\n\n");
      exit(-1);
    }

  if (Options.cid < 1)
    {
      fprintf(stderr, "You have to set CID to a positive value: %i\n",
	      Options.cid);
      fprintf(stderr, "Use option -C to set the CID.\n\n");
      exit(-1);
    }

  /* read in the database options from a file, the servsock.conf file
   * will work here also, additional options will be ignored.
   */
  ReadConfig(&Options);

  if (Options.disabletags && Options.followtags)
    {
      fprintf(output,"Following tagged packets is disabled!\n");
      Options.followtags=0;
    }

  /* Special case for PostgreSQL, if socket is "NULL" set it to real
   * zero value
   */
  if (!strncmp(Options.SocketName,"NULL",4)) Options.SocketName[0]='\0';

  /* Open database connection and do some checks */
  if( (res=DatabaseInit(&Options)) ) 
  {
    if (res!=2) Disconnect(&Options);
    exit(1);
  };

  /* fetch the payload information based on sid and cid from database */
  if(GetPacket(&Options))
    {
      fprintf(stderr,"An error occured in GetPacket()-> Exiting.\n");
      Disconnect(&Options);
      exit(1);
    }


#ifdef DEBUG
  fprintf(output,"pcap_header : <%s>\n",pbuffer);
  fprintf(output,"data_header : <%s>\n",tbuffer);
  if (dbuffer != NULL)
    {
      fprintf(output,"data_payload: <%s>\n",dbuffer);
    }
  else
    {
      fprintf(output,"data_payload: no payload\n");
    }
#endif

  /* finally build the pcap packet and write it to a file or stdout */

  /* first we create the pcap header */
  if (BuildPacketHeader(&Options))
    {
      fprintf(stderr, "An error occured in BuildPacketHeader -> Exiting.\n");
      exit(1);
    }
  
  /* next we append at least one packet */

  if (AppendPacket(&Options))
    {
      fprintf(stderr, "An error occured in AppendPacket -> Exiting.\n");
      exit(1);
    }

  /* Now we should use try to add tagged packets */
  if (Options.followtags)
    {
      if (Verbose)
       {
         fprintf(output,"Following tags\n");
       }
      while(Options.followtags)
	{
	  Options.cid=Options.nextcid;
          res=GetPacket(&Options);
	  if (res==2)
            {
              fprintf(stderr,"No packet information available to rebuild "
                             "-> Skipping packet.\n");
              continue;
            }
          else if(res)
	    {
	      fprintf(stderr,"An error occured in GetPacket()-> Exiting.\n");
	      Disconnect(&Options);
	      exit(1);
	    }

#ifdef DEBUG
	  fprintf(output,"pcap_header : <%s>\n",pbuffer);
	  fprintf(output,"data_header : <%s>\n",tbuffer);
	  fprintf(output,"data_payload: <%s>\n",dbuffer);
#endif
	  AppendPacket(&Options);
	}
    }    

  /* At this point we have all data from the database */
  Disconnect(&Options);

  if (ClosePacket(&Options))
    {
      fprintf(stderr, "An error occured in ClosePacket -> Exiting.\n");
      exit(1);
    }
  
  exit(0);
}


/* 
 * Function: CmdLine
 *
 * Purpose: Set or modify some values needed for startup
 *
 * Arguments: Commandline argv and number of elements argc
 *
 * Returns: Nothing, we set only global values
 *
 * Author: DG
 *
 */

void 
CmdLine(int argc, char* argv[], GPoptions *opt) 
{
  int o, size;
  char options[]="aC:c:dhIl:m:S:tvVw:z";

  while((o = getopt(argc, argv, options)) != -1) 
    {
      switch(o)
	{
	case 'v':
	  printf("\nProgram to extract full packet data from the database\n");
	  printf(" and create a pcap file to analyse it with programs like\n");
          printf(" tcpdump or ethereal.\n");
	  printf("\nWritten by Dirk Geschke\n\n");
	  printf("%s\n\n",rcsid);
	  printf("Version %s (snort-%s)\n\n",FLoP_VERSION,VERSION);
	  exit(0);
	  break;

	case 'd':
	  Verbose=1;
	  break;
	    
        case 'a':
          opt->alltagged=1;
	  opt->followtags=1;
          break; 
      
	case 'h':
	  Usage(argv[0]);
	  break;
	  
	case 'I':
	  Info=1;
	  break;

        case 'l':
          size = strlen(optarg) < STRLEN ? strlen(optarg) : STRLEN - 1;
          strncpy(link_layer_prot, optarg, size) ;
          link_layer_prot[size] = '\0';
          fprintf(output, "\nlink_layer_prot = %s\n", link_layer_prot);
          break;

	case 'c':
	 size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	 strncpy(opt->ConfigFile,optarg, size);
	 opt->ConfigFile[size]='\0';
#ifdef DEBUG_CMDLINE
	 fprintf(output,"ConfigFile: %s\n",opt->ConfigFile); 
#endif
	  break;
	  
	case 'C':
	  opt->cid=atoi(optarg);
#ifdef DEBUG_CMDLINE
	  fprintf(output,"Packet-ID: %i\n",opt->cid);
#endif
	  break;  

	case 'S':
	  opt->sid=atoi(optarg);
#ifdef DEBUG_CMDLINE
	  fprintf(output,"Sensor-ID: %i\n",opt->sid);
#endif
	  break;  

	case 't':
	  opt->followtags=1;
#ifdef DEBUG_CMDLINE
          fprintf(output,"Following tags.\n");
#endif
          break;

	case 'z':
	  opt->disabletags=1;
#ifdef DEBUG_CMDLINE
          fprintf(output,"Ignoring related tagged packets.\n");
#endif
          break;

	case 'm':
	 size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	 strncpy(VendorFile,optarg, size);
	 VendorFile[size]='\0';
	 VendorOpt=1;
	 Info=1;
#ifdef DEBUG_CMDLINE
	 fprintf(output,"ConfigFile: %s\n",opt->ConfigFile); 
#endif
	  break;

	case 'V':
	 VendorOpt=1;
	 Info=1;
#ifdef DEBUG_CMDLINE
	 fprintf(output,"Use vendor map file: %u\n",VendorOpt); 
#endif
	  break;

	case 'w':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(opt->DumpFile,optarg, size);
	  opt->DumpFile[size]='\0';
#ifdef DEBUG_CMDLINE
	  fprintf(output,"Working directory for daemon mode: %s\n",opt->DumpFile);
#endif
	  break;  

	default:
	  Usage(argv[0]);
	}
    }
}

/* 
 * Function: Usage
 *
 * Purpose: Display the possible command line arguments
 *
 * Arguments: argv, only needed to know the program name
 *
 * Returns: No return, simply exit...
 *
 * Author: DG
 *
 */

void 
Usage(char *argv)
{
  printf("\nUsage: %s [-ahtvV] [-c ConfigFile] -C PacketCount"
#ifdef DLT_LINUX_SLL
	 " -l linktype"
#endif
         " -m vendor.map -S Sensor-ID [-w DumpFile] \n",argv);
  printf("\n -a              : Include all tagged packets of a revision (activates -t)");
  printf("\n -c ConfigFile   : Configuration file for database access");
  printf("\n -C PacketCount  : Counter of packet related to sensor (cid)");
  printf("\n -d              : Print more status informations");
#ifdef DLT_LINUX_SLL
  printf("\n -l DLT_LINUX_SLL: Use linux cooked mode instead of the "
         "default ethernet 10 MB");
#endif
  printf("\n -m vendor.map   : Specifies which vendor map should be used");
  printf("\n -S Sensor-ID    : To which sensor belongs the PacketCount");
  printf("\n -t              : Dump all related tagged packets");
  printf("\n -V              : Print vendor informations for MAC header");
  printf("\n -w DumpFile     : Where should the data be written to");
  printf("\n -z              : Disable tagged packets");
  printf("\n\n");
  exit(0);
}

/*
 * Function: DatabaseInit(GPoptions *)
 *
 * Purpose: Connects to database, sets/gets the sensor id/name/encoding
 *          and selects the starting cid from database. Should we change
 *          this to scheme v106?
 *
 * Arguments: args => ptr to database parameters
 *
 * Returns: 0 on success
 *
 */

int 
DatabaseInit(GPoptions *data)
{

  /* allocate memory for configuration queries */
  if (select_ptr==NULL)
    {
      select_ptr = (char *) malloc(MAX_QUERY_LENGTH+1);
      insert_ptr = (char *) malloc(MAX_QUERY_LENGTH+1);
    }
  if ( (select_ptr==NULL) || (insert_ptr==NULL) )
	{
	  fprintf(stderr,"DatabaseInit: problem to allocate memory -> Exiting.\n");
	  return(1);
	}

  /* These values are needed for the Database() function */
  if (pbuffer==NULL)
    {
      dbuffer = (char *) malloc ((MAXBUFFSIZE*2)+1);
      tbuffer = (char *) malloc ((MAXBUFFSIZE*2)+1);
      pbuffer = (char *) malloc ((sizeof(struct pcap_pkthdr)*2)+1);
    }
  if ( (dbuffer==NULL) || (tbuffer==NULL) || (pbuffer==NULL))
    {
      fprintf(stderr,"Database: problem to allocate memory -> Exiting.\n");
      return(1);
    }

  /* Connect to database */
  if (Connect(data))
    {
      return(2);
    }

#ifdef DEBUG_DB
  fprintf(output,"Database> Connected to database.\n");
#endif

  if ( data->DBtype == DB_MYSQL) 
     snprintf(select_ptr,MAX_QUERY_LENGTH,"SELECT full_payload FROM `schema`"); 
  else
     snprintf(select_ptr,MAX_QUERY_LENGTH,"SELECT full_payload FROM schema"); 

  if (Select(select_ptr,data) == 0)
    {
      /* Database is not able to store full payload.
       * Print a warning message and a hint how to
       * activate it. Additionally we have to disable
       * the feature and go on (or should we exit?)
       */
      fprintf(stderr,"The database does not contain the full payload.\n");
      return(1);
    }

  return(0);
}

/* Function: Connect(GPoptions * data)
 *
 * Purpose: Database independent function to initiate a database 
 *          connection
 */

int 
Connect(GPoptions * data)
{
  char *cport;
  char *host=NULL; /* defaults to localhost in mysql */
  int port=0;
#ifdef POSTGRES
  char *c;
#endif

  if ((cport=strchr(data->SocketName,':')) != NULL)
     {
       *cport='\0'; cport++;
       host=data->SocketName;
       port=atoi(cport);
     }

#ifdef MYSQL_DB  
  if ( data->DBtype == DB_MYSQL) 
    {
      data->m_sock = mysql_init(NULL);
      if(data->m_sock == NULL)
	{
	  fprintf(stderr,"database: Connection to database '%s' failed\n", 
		 data->DBname);
	  return(2);
	}

      if( mysql_real_connect(data->m_sock, host, data->DBuser,
	       data->DBpassword, data->DBname, port, data->SocketName, 0) == 0)
	{
	  if(mysql_errno(data->m_sock))
	    {
	      fprintf(stderr, "database: mysql_error: %s\n", 
		      mysql_error(data->m_sock));
	      return(1);
	    }
	  fprintf(stderr,"database: Failed to logon to database '%s'\n", 
		  data->DBname);
	  return(1);
	}
    }
#endif /* MYSQL */
#ifdef POSTGRES
    if ( data->DBtype == DB_POSTGRES ) 
      {
        if ( (c=strstr(data->SocketName,".s.PGSQL") ) != NULL)
           {
              /* split path from name */
             *c='\0'; c++; ;  /* c points now to s */
 
             /* find extension, should start after the next dot. */
             cport=strrchr(c,'.'); cport++; 

              /* host now contains the path */
             host=data->SocketName;
           }

	data->p_sock=PQsetdbLogin(host, cport, NULL, NULL,
				 data->DBname, data->DBuser, data->DBpassword);
	if (PQstatus(data->p_sock) == CONNECTION_BAD)
	  {
	    PQfinish(data->p_sock);
	    if (PQerrorMessage(data->p_sock))
	      fprintf(stderr, "database: psql_error: %s\n",
		     PQerrorMessage(data->p_sock));
	    fprintf(stderr, "database: Failed to logon to postgresql database "
		    "'%s'\n", data->DBname);
	    return(1);
	  }
      }
#endif /* POSTGRES */
    return(0);
}


/* Function: Disconnect(GPoptions * data)
 *
 * Purpose: Close connection to database and free 
 *          allocated memory.
 */

void 
Disconnect(GPoptions * data)
{

#ifdef DEBUG_DB
  fprintf(output,"Database> Closing connection to database \"%s\"\n", data->DBname);
#endif
  if (data)
    {
#ifdef MYSQL_DB
      if((data->DBtype==DB_MYSQL) && data->m_sock ) 
	mysql_close(data->m_sock);
#endif 
#ifdef POSTGRES
      if ( (data->DBtype==DB_POSTGRES) && data->p_sock) 
	PQfinish(data->p_sock);
#endif
    }

  /* to be kindly we free some memory... (but exit() will do it too... */
  free(insert_ptr);
  free(select_ptr);
}

/* Function: Insert(char * query, GPoptions * data)
 *
 * Purpose: Database independent function for SQL inserts
 * 
 * Arguments: query (An SQL insert)
 *
 * Returns: 1 if successful, 0 if fail
 */

int 
Insert(char * query, GPoptions * data)
{
    int result = 0;

#ifdef MYSQL_DB
    if ( data->DBtype == DB_MYSQL ) 
      {
        if(!(mysql_query(data->m_sock,query)))
        {
            result = 1;
        }
        else
        {
            if(mysql_errno(data->m_sock))
            {
              fprintf(stderr,"database: mysql_error: %s\nSQL=%s\n", 
                           mysql_error(data->m_sock), query);
            }
        }
      }
#endif /* MYSQL */

#ifdef POSTGRES
    if ( data->DBtype == DB_POSTGRES ) 
      {
	data->p_result = PQexec(data->p_sock,query);
	if (PQresultStatus(data->p_result) == PGRES_COMMAND_OK)
	    {
	      result = 1;
	    }
	else
	  {
	    if (PQerrorMessage(data->p_sock))
	      fprintf(stderr,"database: psql_error: %s\nSQL=%s\n",
		   PQerrorMessage(data->p_sock),query);
	  }

	/* clear result to avoid memory leak */
	PQclear(data->p_result);
      }
#endif /* POSTGRES */
#ifdef DEBUG_DB
    if(result)
       fprintf(output,"Database> (%s) executed\n", query);
    else
       fprintf(output,"Database> (%s) failed\n", query);
#endif

    return result;
}


/* Function: Select(char * query, GPoptions * data)
 *
 * Purpose: Database independent function for SQL selects that 
 *          return a non zero int
 * 
 * Arguments: query (An SQL select)
 *
 * Returns: result of query if successful, 0 if fail
 */

int 
Select(char * query, GPoptions * data)
{
    int result = 0;

    if (Verbose)
      {
	fprintf(output,"query=%s\n",query);
      }

#ifdef MYSQL_DB
    if ( data->DBtype == DB_MYSQL ) 
      {
        if(mysql_query(data->m_sock,query))
	  {
            result = 0;
	  }
        else
	  {
            if(!(data->m_result = mysql_use_result(data->m_sock)))
	      {
                result = 0;
	      }
            else
	      {
                if((data->m_row = mysql_fetch_row(data->m_result)))
		  {
                    if(data->m_row[0] != NULL)
		      {
                        result = atoi(data->m_row[0]);
		      }
		  }
	      }
            mysql_free_result(data->m_result);
	  }
        if(!result)
	  {
            if(mysql_errno(data->m_sock))
	      {
                fprintf(stderr,"database: mysql_error: %s\n", 
		       mysql_error(data->m_sock));
	      }
	  }
      }
#endif /* MYSQL */
#ifdef POSTGRES
    if ( data->DBtype == DB_POSTGRES ) 
      {
#ifdef DEBUG_DB
	printf("query: %s\n",query);
#endif
	data->p_result = PQexec(data->p_sock,query);
	if ( PQresultStatus(data->p_result) == PGRES_TUPLES_OK)
	  {
	    if (PQntuples(data->p_result))
	      result = atoi(PQgetvalue(data->p_result,0,0));
#ifdef DEBUG_DB
	printf("Status: %i\n",result);
#endif
	  }
	else
	  {
	    result=0;
	    if (PQerrorMessage(data->p_sock))
	     fprintf(stderr, "database: psql_error: %s\nSQL=%s\n",
		     PQerrorMessage(data->p_sock),query);
	  }

	/* clear result to avoid memory leak */
	PQclear(data->p_result);
      }
#endif /* POSTGRES */

#ifdef DEBUG_DB
    if(result)
       fprintf(output,"Database> (%s) returned %u\n", query, result);
    else
       fprintf(output,"Database> (%s) failed\n", query);
#endif

    return result;
}

/* Function: Select1(char * query, GPoptions * data)
 *
 * Purpose: Database independent function for SQL selects that 
 *          return a TEXT string
 * 
 * Arguments: query (An SQL select)
 *
 * Returns: result of query.
 */

char *
Select1(char * query, GPoptions * data)
{
    int result = 0;
    char *res=NULL;

    if (Verbose)
      {
	fprintf(output,"query=%s\n",query);
      }

#ifdef MYSQL_DB
    if ( data->DBtype == DB_MYSQL ) 
      {
        if(mysql_query(data->m_sock,query))
	  {
            result = 0;
	  }
        else
	  {
            if(!(data->m_result = mysql_use_result(data->m_sock)))
	      {
                result = 0;
	      }
            else
	      {
                if((data->m_row = mysql_fetch_row(data->m_result)))
		  {
                    if(data->m_row[0] != NULL)
		      {
                        res = strdup(data->m_row[0]);
			if (res==NULL) result=0;
		      }
		  }
	      }
            mysql_free_result(data->m_result);
	  }
        if(!result)
	  {
            if(mysql_errno(data->m_sock))
	      {
                fprintf(stderr, "database: mysql_error: %s\n", 
		       mysql_error(data->m_sock));
	      }
	  }
      }
#endif /* MYSQL */
#ifdef POSTGRES
    if ( data->DBtype == DB_POSTGRES ) 
      {
#ifdef DEBUG_DB
	printf("query: %s\n",query);
#endif
	data->p_result = PQexec(data->p_sock,query);
	if ( PQresultStatus(data->p_result) == PGRES_TUPLES_OK)
	  {
	    if (PQntuples(data->p_result))
              {
		res = strdup ((char *) PQgetvalue(data->p_result,0,0));
	      }

#ifdef DEBUG_DB
	printf("Status: %s\n",res);
#endif
	  }
	else
	  {
	    result=0;
	    if (PQerrorMessage(data->p_sock))
	      fprintf(stderr,"database: psql_error: %s\nSQL=%s\n",
		     PQerrorMessage(data->p_sock),query);
	  }

	/* clear result to avoid memory leak */
	PQclear(data->p_result);
      }

#endif /* POSTGRES */

    return res;
}


/* Function: ReadConfig()
 *
 * Purpose: read configuration files to set at least the parameter
 *          for database access (avoid password on command line)
 *          spaces/tabs at the beginning are removed, spaces at
 *          the end too. But maybe we should use a better parser...
 *
 * Author: DG
 * 
 */

void 
ReadConfig(GPoptions* Options)
{
  FILE *fd = NULL;
  char *token =NULL;
  char *arg=NULL;
  char *val=NULL;
  char line[STRLEN];
  int linenumber=0;
  int len;

  if ( (fd=fopen(Options->ConfigFile,"r")) == NULL ) 
    {
      fprintf(stderr, "Can not open config file \"%s\": %s\n", 
	      Options->ConfigFile, strerror(errno));
      exit(1);
    }
  while((fgets(line, STRLEN, fd))!=NULL ) 
    { 
      linenumber++;

      /* remove end of line */
      if ( (token = strchr (line,'\n')) ) *token='\0';

      /* skip if line starts with a comment: if no comment sign */
      /* is found we are sure that we can derefence the pointer */
      /* before any comment sign!*/

      if ( line[0]=='#' ) continue;

      /* remove any comments at end of line and if a comment */
      /* sign is escaped: remove backslash from string!      */
      token = line;
      while ( (token = strchr(token,'#')) && (*(token-1)=='\\' ) )
      {
         val=token;
         arg=token-1;
	 while (*arg) {
		 *arg=*token;
		 arg++;
		 token++;
	 }
	 token=val;
      }

      /* Maybe token is pointing now to a real comment sign */
      if (token && *token=='#') *token='\0';

      /* next if line empty (or only a comment) */
      if (line[0]=='\0') continue;

      /* Up to ":" or "=" we get the argument */
      if ( (strstr(line,":") == NULL) && (strstr(line,"=") == NULL) )
        {
           fprintf(stderr,"\nIgnoring unknown entry in config file: \n");
           fprintf(stderr,"\n> %s (missing : or = ?)\n\n",line);
           continue;
        }

      arg = strtok(line,":=");
      /* second argument is the value, remove tabs */
      val = strtok(NULL,"\0");
      /* remove leading spaces or tabs */
      while ( arg[0]==' ' || arg[0]=='\t') arg++;
      while ( val[0]==' ' || val[0]=='\t') val++;
      /* remove spaces at end of value */
      while ( (strlen(arg) > 1) && (arg[strlen(arg)-1] == ' ') )
		    arg[strlen(arg)-1]='\0';
      while ( (strlen(val) > 1) && (val[strlen(val)-1] == ' ') )
		    val[strlen(val)-1]='\0';
#ifdef DEBUG_READCONFIG
      fprintf(output,"Line: %i Argument: <%s> Value: <%s>\n",linenumber,arg,val);
#endif
      ArgCheck(val);

      len=strlen(val);
      if(len > STRLEN-2) 
	{
	  fprintf(stderr, "Value in config file too long: %s\n",val);
	  exit(1);
	}

       if ( !strncasecmp(arg,"reference",9) )
        {
           if(val)
           {
	     Options->followtags=1;
	   }
	   continue;
        } 

      if ( !strncasecmp(arg,"Info",5) )
	{
	  if (Info) continue; /* already set via command line */

	  Info=atoi(val);
	  continue;
	}

      if ( !strncasecmp(arg,"vendormap",9) )
	{
	  if( VendorOpt) continue; /* already set via command line */

	  strncpy(VendorFile,val,strlen(val));
	  VendorFile[len]='\0';
	  VendorOpt=1;
	  continue;
	}

      if ( !strncasecmp(arg,"vendorinfo",10) )
	{
	  if( VendorOpt) continue; /* already set via command line */
	  if (atoi(val)) VendorOpt=1;
	  continue;
	}

       if ( !strncasecmp(arg,"dbuser",6) )
        {
           strncpy(Options->DBuser,val,len);
	   Options->DBuser[len]='\0';
	   continue;
        } 

      if ( !strncasecmp(arg,"dbpassword",10) )
	{
	      strncpy(Options->DBpassword,val,len);
	      Options->DBpassword[len]='\0';
	      continue;
	}

      if ( !strncasecmp(arg,"dbname",6) )
	{
	      strncpy(Options->DBname,val,strlen(val));
	      Options->DBname[len]='\0';
	      continue;
	}

      if ( !strncasecmp(arg,"socketname",10) )
	{
	  if (len > UNIX_PATH_MAX)
	    {
	      fprintf(stderr,"Length of socketname too large!\n");
	      fprintf(stderr,"Maximum length: %u\n",UNIX_PATH_MAX);
	      fprintf(stderr,"Actual length : %u\n",len);
	      fprintf(stderr,"Exiting.\n");
	      exit(1);
	    }
	  strncpy(Options->SocketName,val,len);
	  Options->SocketName[len]='\0';
	  continue;
	}

      if ( !strncasecmp(arg,"dbtype",7) )
	{
#ifdef MYSQL_DB
	  if (!strncasecmp(val,"MYSQL",5))
	    { 
#ifdef DEBUG_DB
	      fprintf(output,"Database type: MySQL\n");
#endif
	      Options->DBtype=DB_MYSQL;
	    }
#endif
#ifdef POSTGRES
	  if (!strncasecmp(val,"Postgres",8))
	    { 
#ifdef DEBUG_DB
	      fprintf(output,"Database type: Postgres\n");
#endif
	      Options->DBtype=DB_POSTGRES ;
	    }
#endif
	if (Options->DBtype==0)
          {
             fprintf(output,"WARNING: DBtype not set (%s)!\n",val);
             fprintf(output,"Did you compile in support for it?\n");
          }
	  continue;
	}

      /* print warning to stderr to avoid conflicts with stdout */
      fprintf(stderr, "WARNING: unknown option in line %i: %s\n", linenumber,
	      arg);
      
    }

  fclose(fd);
}


/* 
 *  Function: ArgCheck
 *
 *  Purpose: Check for occurence of wrapper around the arguments
 *           in the config file. We remove ' and " if they are at
 *           the beginning and the end.
 *
 *  Arguments: the extracted value as string
 *
 *  Author: DG
 *
 */

void 
ArgCheck(char *arg)
{
  char Value[STRLEN];
  int len;
  
  len=strlen(arg);

  if (len < 2) return;
#ifdef DEBUG_READCONFIG
  fprintf(output,"Arg: %s, Len: %i\n",arg,len);  
#endif

  if ( ((index(arg,'"')==arg) && (rindex(arg,'"')==arg+len-1)) ||
       ((index(arg,'\'')==arg) && (rindex(arg,'\'')==arg+len-1)) )
   {
       strncpy(Value,arg+1,len-2);
       /* dont't forget to end the string... */
       Value[len-2]='\0';
#ifdef DEBUG_READCONFIG
       fprintf(output,"Value: %s\n",Value);
#endif
       strcpy(arg,Value);
   }
  return;
}


/* 
 * Function: GetPacket
 *
 * Purpose: Fetches the information to rebuild a network pcap packet
 *          from the database based on sid and cid
 *
 * Arguments: GPoptions, containing database information, sid and cid
 *
 * Returns: 0 on success, 1 on error
 *
 * Author: DG
 *
 */

int
GetPacket(GPoptions *Options)
{
  char *result;
  int nodataheader=0;

  /* First get the encoding */
  snprintf(select_ptr,MAX_QUERY_LENGTH-1,"SELECT encoding FROM sensor "
	   "WHERE sid = '%i'",Options->sid);
 
  Options->DBencoding = Select(select_ptr,Options);
#ifdef DEBUG_GP
  fprintf(output,"Encoding: %i\n",Options->DBencoding);
#endif
  
  if ((Options->DBencoding != ENCODING_HEX) &&
      (Options->DBencoding != ENCODING_BASE64))
    {
      fprintf(stderr,"Actually only HEX and Base64 encoding is supported\n");
      return(1);
    }
  
  /* Next get the reference */
  if (Options->followtags)
    {
      snprintf(select_ptr, MAX_QUERY_LENGTH-1, "SELECT reference FROM event "
              "WHERE sid = '%i' AND cid = '%i'",Options->sid,Options->cid);
 
      Options->reference = Select(select_ptr,Options);
      if (Options->reference == 0)
        {
          fprintf(stderr, "\nNo reference found, cannot follow tags.\n");
          Options->followtags=0;
        }
      else
	{
	  if (Verbose)
	    {
	      fprintf(output,"Reference = %i\n", Options->reference);
	    }

	  /* Get the CID of the first packet starting the tagged list */
          if (Options->alltagged)
	    {
              Options->alltagged=0;
	      if (Verbose)
		{
		  fprintf(output,"Given CID = %i\n", Options->cid);
		}
	      snprintf(select_ptr, MAX_QUERY_LENGTH-1, "SELECT cid FROM event "
		       "WHERE sid = '%i' AND reference = '%i' ORDER BY cid "
		       "LIMIT 1",
		       Options->sid, Options->reference);

	      Options->cid = Select(select_ptr,Options);
	      if (Verbose)
		{
		  fprintf(output,"New CID = %i\n",Options->cid);
		}
	    }

          snprintf(select_ptr, MAX_QUERY_LENGTH-1, "SELECT cid FROM event "
                   "WHERE sid = '%i' AND cid > '%i' AND reference = '%i' "
		   "ORDER BY cid LIMIT 1",
                   Options->sid, Options->cid, Options->reference);

          Options->nextcid = Select(select_ptr,Options);

          if (Options->nextcid==0) 
            {
	      if (Verbose)
		{
		  fprintf(output,"No more tagged packets found after CID=%i.\n",
			 Options->cid);
		}
              Options->followtags=0;
              Options->reference=0;
            }
          else
            {
	      if (Verbose)
		{
		  fprintf(output,"Next CID=%i.\n",Options->nextcid);
		}
            }
        }
    }

  /* Next get the pcap_header */
  snprintf(select_ptr, MAX_QUERY_LENGTH-1, "SELECT pcap_header FROM data "
	   "WHERE sid = '%i' AND cid = '%i'",Options->sid,Options->cid);
 
  result = Select1(select_ptr,Options);
  if (result==NULL) 
    {
      /* We don't have a pcap_header, so skip rebuilding */
      fprintf(stderr,"No pcap header, we can't rebuild the network packet.\n");
      return(2);
    }

  if (Options->DBencoding == ENCODING_BASE64) 
   { 
       if ( base64tohex(result) < 0 ) return(1);
   }

  if (strlen(result) > (sizeof(struct pcap_pkthdr)*2))
    {
      fprintf(stderr,"Error in pcap_header, answer is too large: %i (%i)!\n",
	     (int) strlen(result),sizeof(struct pcap_pkthdr)*2);
      return(1);
    }
#ifdef DEBUG
  fprintf(output,"result: <%s>, len %i\n",result,(int) strlen(result));
#endif
  snprintf(pbuffer,(sizeof(struct pcap_pkthdr)*2)+1,"%s",result);

  free(result);

  /* Next get the full header of the network packet */
   snprintf(select_ptr, MAX_QUERY_LENGTH-1, "SELECT data_header FROM "
	    "data WHERE sid = '%i' AND cid = '%i'",Options->sid,Options->cid);
 
  result = Select1(select_ptr,Options);

  if (result==NULL) 
    {
      /* We don't have a pcap_header, so skip rebuilding */
      fprintf(stderr,"No data header, maybe we can't rebuild the packet.\n");
      nodataheader=1;
    }

  if (result && (strlen(result) > MAX_QUERY_LENGTH-1))
    {
      fprintf(stderr,"Please increase 2 * MAX_QUERY_LENGTH to at least %i\n",
	     (int) strlen(result));

      free(result);

      return(1);
    }

#ifdef DEBUG_GP
  fprintf(output,"result: <%s>, len %i\n",result,(int)strlen(result));
#endif

  snprintf(tbuffer,MAX_QUERY_LENGTH*2-1,"%s",result);

  if (Options->DBencoding == ENCODING_BASE64)
   {
       if ( base64tohex(tbuffer) < 0 ) return(1);
   }

  if (result) free(result);

  /* Finally get the payload of the packet */

  snprintf(select_ptr, MAX_QUERY_LENGTH-1, "SELECT data_payload FROM data WHERE sid = '%i' AND cid = '%i'",Options->sid,Options->cid);
  
  result = Select1(select_ptr,Options);
  if (result==NULL) 
    {
      if (Verbose)
         fprintf(stderr,"Warning: No high layer payload.\n");
      /* if we have neither a header nor a payload we should 
       * skip this packet
       */
      if (nodataheader) 
	{
	  fprintf(stderr,
		    "Neither data_header nor data_payload are available!\n");
	  return(2);
	}
    }

  /* Final test, should never happen */
  if (result && (strlen(result) > MAXBUFFSIZE*2-1))
    {
      fprintf(stderr,"Please increase MAXBUFFSIZE to at least %i\n",
	     (int) strlen(result)/2+1);

      free(result);

      return(1);
    }
#ifdef DEBUG_GP
  fprintf(output,"result: <%s>, len %i\n",result, (int) strlen(result));
#endif

  snprintf(dbuffer,MAXBUFFSIZE*2-1,"%s",result);
  if (result) free(result);
  
  if (Options->DBencoding == ENCODING_BASE64)
   {
       if ( base64tohex(dbuffer) < 0 ) return(1);
   }

  return(0);
}


/* 
 * Function: BuildPacketHeader
 *
 * Purpose: builds a network pcap packet heaer and writes it to a file
 *
 * Arguments: GPoptions, containing database information, sid and cid
 *
 * Returns: 0 on success, 1 on error
 *
 * Author: DG
 *
 */

int 
BuildPacketHeader(GPoptions* Options)
{
  int snaplen;
  struct pcap_file_header hdr;
  
  /* First open the file if possible or stdout */

  if (Options->DumpFile[0]=='-' && Options->DumpFile[1]=='\0')
    {
      /* Dump to stdout */
      dumpfd=stdout;
 
      /* print everything to stderr instead of stdout */
      output=stderr; 
    }
  else
    {
      dumpfd=fopen(Options->DumpFile,"w");
    }
     
  /* Check file descriptor */
  if ( dumpfd == NULL ) 
    {
      fprintf(stderr, "Can not open pcap file for writing \"%s\": %s\n", 
	     Options->DumpFile, strerror(errno));
      return(1);
    }

  /* snaplen is len of payload plus header */
  /* for HEX we have to divide by two -> two HEX characters
   * represent one binary byte.
   */

  snaplen=(strlen(tbuffer)+strlen(dbuffer))/2;

  /* Build pcap file header */
  hdr.magic=0xa1b2c3d4;
  hdr.version_major= PCAP_VERSION_MAJOR;
  hdr.version_minor= PCAP_VERSION_MINOR;
  hdr.thiszone=0;                        /* No offset due to timezone */
  hdr.snaplen=snaplen;
  hdr.sigfigs=0;
  hdr.linktype=DLT_EN10MB;

  if (!strcmp(link_layer_prot, "DLT_LINUX_SLL"))
    {
#ifdef DLT_LINUX_SLL
      hdr.linktype = DLT_LINUX_SLL;
#else
      fprintf(stderr,"DLT_LINUX_SLL not supported on this platform, ");
      fprintf(stderr,"using DLT_10ENB instead.");
#endif
    }


  /* write the pcap file header */
#ifdef HAVE_PCAP_HEADER
  pd = pcap_open_dead(hdr.linktype, snaplen);
  if (pd == NULL)
    {
      fprintf(stderr, "pcap_open_dead returned NULL!\n");
            fclose(dumpfd);
      return(1);
    }
  p = pcap_dump_fopen(pd, dumpfd);
  if (p == NULL)
    {
      fprintf(stderr, "pcap_dump_fopen: %s\n", pcap_geterr(pd));
      fclose(dumpfd);
      return(1);
    }
#else
  if (fwrite((char*) &hdr, sizeof(hdr),1,dumpfd) != 1)
    {
      fprintf(stderr,"Problem writing header file.\n");
      fclose(dumpfd);
      return(1);
    }
#endif
 
  return(0);
}

/* 
 * Function: AppendPacket
 *
 * Purpose: builds a network pcap packet and appends it to a file
 *
 * Arguments: GPoptions, containing database information, sid and cid
 *
 * Returns: 0 on success, 1 on error
 *
 * Author: DG/AJB
 *
 */

int 
AppendPacket(GPoptions* Options)
{
  int snaplen;
  DG_pcap_pkthdr phdr;
  char tv[17];
  static char packet[MAXBUFFSIZE];
  char * packet_ptr;
  char * p_ptr;
  char * d_ptr;
  char * t_ptr;
  char * t;
  char ts[4];
  int i,j,k;
  int sizetime;
  unsigned int packet_type, addr_len;

  sizetime=sizeof(struct timeval);

#ifdef DEBUG_AP
  fprintf(output,"Sizetime: %i\n",sizetime);
#endif

  /* snaplen is len of payload plus header */
  /* for HEX we have to divide by two -> two HEX characters
   * represent one binary byte.
   */

  /* Set start points of buffer */
  p_ptr=pbuffer;
  t_ptr=tbuffer;
  d_ptr=dbuffer;

  snaplen=(strlen(tbuffer)+strlen(dbuffer))/2;

  if (Info)
    {
      t=t_ptr;
#ifdef DEBUG
      printf("tbuffer: <%s> (%i)\n",tbuffer, (int) strlen(tbuffer));
      printf("dbuffer: <%s> (%i)\n",dbuffer, (int) strlen(dbuffer));
#endif
      if (!strcmp(link_layer_prot, "DLT_LINUX_SLL"))
      {
       /* cf. libpcap-0.6.2/sll.h:  struct sll_header */
         packet_type = hextoint(t, 2) * 256 + hextoint(t+2, 2);
         fprintf(output,"\n Pkt type  : %i ", packet_type); /* network byte order */
         switch (packet_type)
         {
                 case 0:
                         fprintf(output, "(= host)\n");
                         break;
                 case 1:
                         fprintf(output, "(= broadcast)\n");
                         break;
                 case 2:
                         fprintf(output, "(= multicast)\n");
                         break;
                 case 3:
                         fprintf(output, "(= to other host)\n");
                         break;
                 case 4:
                         fprintf(output, "(= outgoing)\n");
                         break;
                 default:
                         fprintf(output, "( = unknown)\n");
         }
         t += 4;
         fprintf(output," Addr type : %u (= 0x%.04X)\n", 
                 hextoint(t,2) * 256 + hextoint(t+2, 2), 
                 ( hextoint(t,2) * 256 + hextoint(t+2,2) ) ); /* network byte order */
         t += 4;
         addr_len = hextoint(t, 2) * 256 + hextoint(t + 2, 2);
         fprintf(output," Addr len  : %i \n", addr_len);
         t += 4;
         fprintf(output," Src-MAC   : "); /* NO network byte order! */
         for (k = 0; k < addr_len; k++)
         {
                 fprintf(output, "%c%c", t[k * 2], t[k * 2 + 1]);
                 if (k + 1 < addr_len)
                 {
                         fprintf(output, ":");
                 }
         }
         fprintf(output, "\n");
         if (VendorOpt)
              fprintf(output," Vendor    : %s\n",GetVendor(t));
         t +=16; /* the latter field always takes 8 Bytes */
         fprintf(output," Protocol  : %c%c%c%c\n",t[0],t[1],t[2],t[3]);
         fprintf(output," IP-Version: %c\n",t[4]);

      } else {         /* means DLT_EN10MB;
                        * With ethernet
                        * Dest comes first, then Src;
                        * see also RFC 826, RFC 2464 (sub 3.);
                        * see also snort-2.4.3/src/decode.h */
      fprintf(output,"\n Dest-MAC  : %c%c:%c%c:%c%c:%c%c:%c%c:%c%c\n",
	      t[0],t[1],t[2],t[3],t[4],t[5],t[6],t[7],t[8],t[9],t[10],t[11]);
      if (VendorOpt)
	fprintf(output," Vendor    : %s\n",GetVendor(t));
      t+=12;
      fprintf(output," Source-MAC: %c%c:%c%c:%c%c:%c%c:%c%c:%c%c\n",
	      t[0],t[1],t[2],t[3],t[4],t[5],t[6],t[7],t[8],t[9],t[10],t[11]);
      if (VendorOpt)
	fprintf(output," Vendor    : %s\n",GetVendor(t));
      t+=12;
      fprintf(output," Protocol  : %c%c%c%c\n",t[0],t[1],t[2],t[3]);
      fprintf(output," IP-Version: %c\n",t[4]);
        /* end of DLT_EN10MB */
      } /* if (!strcmp(link_layer_prot, "DLT_LINUX_SLL")) */

      if (t[4]=='4')
	{
	  t+=5;
	  snprintf(ts,2,"%s",t);
	  fprintf(output," IP Hdr Len: %u\n",atoi(ts)*4);
	  t++;
	  fprintf(output," Flags     : 0x%c%c\n",t[0],t[1]);
	  t+=2;
	  fprintf(output," Ttl Length: %u\n",hextoint(t,2)*256+hextoint(t+2,2));
          t+=4;
          fprintf(output, " Id        : %u (= 0x%.04X)\n", 
                  hextoint(t,2) * 256 + hextoint(t+2,2), 
                  hextoint(t,2) * 256 + hextoint(t+2,2)); /*network byte order*/
          t+=8; /* Fragmentation skipped */

	  fprintf(output," TTL       : %u\n",hextoint(t,2));
	  t+=2;
	  fprintf(output," Protocol  : %u\n",hextoint(t,2));
	  t+=6; /* Checksum skipped */
	  fprintf(output," Source IP : %u.%u.%u.%u\n", hextoint(t,2),
		  hextoint(t+2,2), hextoint(t+4,2), hextoint(t+6,2));
	  t+=8;
	  fprintf(output," Dest IP   : %u.%u.%u.%u\n\n", hextoint(t,2),
		  hextoint(t+2,2), hextoint(t+4,2), hextoint(t+6,2));
	}
    }

  tv[sizetime]='\0'; 
  /* create pcap packet header, this consists of 4 fields a 4 byte */
  /* We have to take care of endianess! */
 
  memcpy(tv,p_ptr,sizetime);

  if ((endian == DG_LITTLE_ENDIAN) && (sizetime == 16)) 
    {
        /* Limit to 32 bit 
	 * This is a special case on little endian systems
	 * with 64 bit. Since the year 2038 is far away we
	 * can safely assume that the higher numbers are all
	 * zero. By limiting tv to 8 bytes we have only the
	 * lower part on a little endian system. Thus the
	 * strtoul function works fine. Otherwise we have to
	 * use stroull and then swap the first and second 32 
	 * bits in order to have htonl working.
	 *
	 * NOTE: On 64-bit big endian system this is not a
	 * problem, then the zeroes are leading and therefore
	 * strtoul works by ingoring the leading zeroe and
	 * htonl() is only an empty function.
	 * */
	  tv[8]='\0';
    }

#ifdef DEBUG_AP
  fprintf(output,"sec: %s\n",tv);
#endif

  phdr.ts.tv_sec=htonl(strtoul(tv,(char **)NULL,16));

  p_ptr+=sizetime;
  memcpy(tv,p_ptr,sizetime);

  if ((endian == DG_LITTLE_ENDIAN) && (sizetime == 16)) 
    {
        /* Limit to 32 bit 
	 * This is a special case on little endian systems
	 * with 64 bit. Since the year 2038 is far away we
	 * can safely assume that the higher numbers are all
	 * zero. By limiting tv to 8 bytes we have only the
	 * lower part on a little endian system. Thus the
	 * strtoul function works fine. Otherwise we have to
	 * use stroull and then swap the first and second 32 
	 * bits in order to have htonl working.
	 *
	 * NOTE: On 64-bit big endian system this is not a
	 * problem, then the zeroes are leading and therefore
	 * strtoul works by ingoring the leading zeroe and
	 * htonl() is only an empty function.
	 * */
	  tv[8]='\0';
    }

#ifdef DEBUG_AP
  fprintf(output,"usc: %s\n",tv);
#endif

  phdr.ts.tv_usec=htonl(strtoul(tv,(char **)NULL,16));

  /* Limit again to 32 bit */
  tv[8]='\0';

  p_ptr+=sizetime;
  memcpy(tv,p_ptr,8);
#ifdef DEBUG_AP
  fprintf(output,"cap: %s\n",tv);
#endif
  phdr.caplen=htonl(strtoul(tv,(char **)NULL,16));

  p_ptr+=8;
  memcpy(tv,p_ptr,8);
#ifdef DEBUG_AP
  fprintf(output,"len: %s\n",tv);
#endif
  phdr.len=htonl(strtoul(tv,(char **)NULL,16));

#ifdef DEBUG_AP
  fprintf(output,"len, caplen: %i, %i\n",phdr.len,phdr.caplen);
  fprintf(output,"snaplen    : %i\n",snaplen);  
#endif 

  if (snaplen != phdr.caplen)
    {
      fprintf(stderr,"WARNING: caplen and snaplen different: %i / %i\n",
	      snaplen,phdr.caplen);
    }

#ifndef HAVE_PCAP_HEADER
  /* write the pcap  header */

  if (fwrite((char*) &phdr, sizeof(phdr),1,dumpfd) != 1)
    {
      fprintf(stderr,"Problem writing header file.\n");
      fclose(dumpfd);
      return(1);
    }
#endif
 
  tv[2]='\0';
  i=0;
  j=strlen(tbuffer);
  packet_ptr=packet;

  /* copy header to packet, change HEX to BIN */
  while (i<j)
    {
      memcpy(tv,t_ptr,2);
      packet_ptr[0]=(char )strtoul(tv,(char **)NULL,16);
      packet_ptr++; 
      t_ptr+=2; 
      i+=2;
    }

  i=0;
  j=strlen(dbuffer);

  /* copy header to packet, change HEX to BIN */
  while (i<j)
    {
      memcpy(tv,d_ptr,2);
      packet_ptr[0]=(char )strtoul(tv,(char **)NULL,16);
      packet_ptr++; 
      d_ptr+=2; 
      i+=2;
    }

#ifdef HAVE_PCAP_HEADER
  pcap_dump((u_char *)p, &phdr, packet);
#ifdef HAVE_PCAP_DUMP_FLUSH
  pcap_dump_flush(p);
#endif
#else /* HAVE_PCAP_HEADER */
  /* write the packet to the file */
  if (fwrite(packet, snaplen, 1, dumpfd) != 1)
    {
      fprintf(stderr,"Problem writing header file.\n");
      fclose(dumpfd);
      return(1);
    }
#endif /* HAVE_PCAP_HEADER */

  Count++;

#ifdef DEGUG_BP 
  fprintf(output,"Done.\n");
#endif

  return(0);
}
/* 
 * Function: ClosePacket
 *
 * Purpose: closes a network pcap file
 *
 * Arguments: GPoptions, containing database information, sid and cid
 *
 * Returns: 0 on success, 1 on error
 *
 * Author: DG/AJB
 *
 */

int
ClosePacket(GPoptions* Options)
{
  if (Verbose)
    {
      fprintf(output,"Packets written %i\n",Count);
    }
#ifdef HAVE_PCAP_HEADER
  pcap_dump_close(p);
#else
  fclose(dumpfd);
#endif
  return(0);
}

/*
 * Function: hextoint
 *
 * Purpose: converts a string to an integer value
 *
 * Arguments: string to convert and number of characters to use
 *
 * Returns: unsingned integer
 *
 * Author: DG
 *
 */

unsigned int 
hextoint(char *t, int len)
{
  unsigned int i,j=0,k=0;

  if(len<1) return(0);

  for (i=0; i<len; i++)
    {
      j=t[0]-'0'; 
      if (j > 9) j-=7;
      k=k+j*16*(len-1-i);
      /*      printf("j: %i, t: %c, k: %i\n",j,t[0],k);*/
      t++;
    }

  return(k+j);
}

/*
 * Function: GetVendor
 *
 * Purpose: Finds the vendor of an ethernet card via the mac address
 *
 * Arguments: string containing the MAC address
 *
 * Returns: string with vendor information
 *
 * Author: DG
 *
 */

char * 
GetVendor(char *t)
{
  static char Vendor[STRLEN]="(no vendor information available)";
  char line[STRLEN];
  int linenumber=0;
  int i;
  char * token;
  FILE * fd;

  if ( (fd=fopen(VendorFile,"r")) == NULL ) 
    {
      fprintf(stderr,"Can't find vendor file: %s\n",VendorFile);
      Vendor[0]='\0';
      VendorOpt=0;
      return Vendor;
    }
  while((fgets(line,STRLEN,fd)) != NULL)
    {
      linenumber++;

      /* remove end of line */
      if ( (token = strchr (line,'\n')) ) *token='\0';

      /* skip if line starts with a comment: if no comment sign */
      /* is found we are sure that we can derefence the pointer */
      /* before any comment sign!*/

      if ( line[0]=='#' ) continue;

      /* First sign should be par of a hex range */ 
      if ( line[0] < '0' || line[0] > 'F') continue;
      if ( line[0] > '9' && line[0] < 'A') continue;
      if ( !strncmp(t,line,6))
	{
	  /* We found the mac address */
          /* remove leading spaces */
          i=6;
          while (line[i]==' ') { i++; }
          strcpy(Vendor,line+i);
	  break;
	}
    }

  fclose(fd);
  return(Vendor);
}


/*
 * Function: CheckEndian
 *
 * Purpose: check endian type of the computer
 *
 * Arguments: none
 *
 * Returns: LITTLE_ENDIAN or BIG_ENDIAN as an integer
 *
 * Author: DG
 *
 */

char 
CheckEndian()
{
  union 
  { 
    struct { u_int8_t b1, b2; } Byte;
    u_int16_t Word;
  } Endian;

 Endian.Word=0x1234;

 if (Endian.Byte.b1 == 0x12) return(DG_BIG_ENDIAN);
 else  return(DG_LITTLE_ENDIAN);
}


/*
 * Function: base64decode
 *
 * Purpose: decode base64 encoded data back to binary data
 *
 * Arguments: pointer to the string which should be converted
 *
 * Returns: number of binary bytes or -1 if an error occurs
 *          the binary data is stored in the same buffer since
 *          it should fit. binary data is 3/4 of base64 data.
 *
 * Author: DG
 *
 */
int 
base64decode(unsigned char *line)
{
   unsigned char  p;
   unsigned short c;
   int i,j,count,len;
   /* this lists contains the translation from ascii (-43) to  
    * Base64 values, note 99 indicates an error. We use unsigned
    * short to aviod thinking about signed values if we use >>.
    */
   static const unsigned char rb[123] = {
    62, 99, 99, 99, 63, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 99, 
    99, 99, 99, 99, 99, 99,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 
    10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 
    99, 99, 99, 99, 99, 99, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 
    36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51 };

   len=strlen(line);

   if (len==0) return(0);

   j=0; count=0;

   for(i=0; i<len; i++)
    {
      p=line[i];
      if ((count%72==0) && p=='\n')
        {
          /* newline */ 
          continue;
        }

    
      /* is it a padding sign? Then we are ready...*/
      if (p == '=') 
        {
          line[j]='\0';
          len=count*3/4;
          if (len>j+1) 
            { 
              fprintf(stderr,"Warning: base64 decoded string too short!\n");
              return(j);
            }
          return(len);
        }

      /* is this a base64 allowed character? */
      if ( p > 122 || p < 43 ) 
        {
           fprintf(stderr,"base64decode> illegal character no %i: <%c>\n", i, 
                   line[i]);
           return (-1);
        }
 
      /* subtract 43 from ascii value so we can save some memory */
      p=p-43;

      /* 99 is not base64...*/
      if (rb[p]==99) 
        {
           fprintf(stderr,"base64decode> illegal character: <%c>\n",line[i]);
           return (-1);
        }
 
      /* get value of base64 character */
      c=rb[p];

#ifdef DEBUG
      printf("p: %c, c:%i\n",p+43,c);
#endif

      switch (i%4)
       {
         case 0:
		 line[j]= c << 2;
		 break;

         case 1:
		 line[j] |= c >> 4; 
		 j++;
		 line[j] = c << 4;
		 break;

         case 2:
		 line[j] |= c >> 2;
		 j++;
		 line[j] = c << 6;
		 break;

         case 3:
		 line[j] |= c;
		 j++;
		 break;

       }
    count++;
    }
   /* append a \0, maybe it is an ascii string and someone wants
    * to print it to the screen? 
    */
   if (j<len) line[j+1]='\0';

   /* calculate length, 4 base64 values are 3 binary, so the new 
    * length should be 3/4 of the old one.
    */
   len=count*3/4;
   if (len>j+1) 
     { 
       fprintf(stderr,"Warning: base64 decoded string too short!\n");
       return(j);
     }
   return(len);
}


/*
 * Function: bin2hex
 *
 * Purpose: transforms binary data to hex encoded string
 *
 * Arguments: pointer to the data which should be converted and
 *            length of data (the data could contain binary zero)
 *
 * Returns: -1 if an error occurs, otherwise 0
 *          the converted data is stored 
 *
 * Author: DG
 *
 */
int
bin2hex(unsigned char *data,int length)
{
    u_char conv[] = "0123456789ABCDEF";
    u_char *index;
    u_char *end;
    u_char *ridx;
    static u_char * buffer = NULL;

    index = data;
    end = index + length;

    if ((length+1)*3/4 > MAXBUFFSIZE)
      {
        fprintf(stderr,"bin2hex: data to big -> %i",length);
        return(-1);
      }

#ifdef DEBUG
    printf("bin2hex> buffer %p\n",buffer);
    printf("bin2hex> data at %p, len: %u\n", data,length);
#endif

    if (buffer == NULL)
      {
         buffer = (char *) malloc((MAXBUFFSIZE*2)+1);
         if ( buffer == NULL )
           {
             fprintf(stderr,"bin2hex: Problem to allocate memory.\n");
             return(-1);
           }
      }

    /* maybe there is no data to convert? */
    if (length < 1)
      {
         snprintf(data,MAXBUFFSIZE*2,"{No data}");
         return(0);
#ifdef DEBUG
         printf("Fasthex> length zero, returning: <%s>\n",buffer);
#endif
      }

    ridx = buffer;

    while(index < end)
    {
        *ridx++ = conv[((*index & 0xFF)>>4)];
        *ridx++ = conv[((*index & 0xFF)&0x0F)];
        index++;
    }

    *ridx='\0';

    strncpy(data,buffer,length*2+1);

    return(0);
}


/*
 * Function: base64tohex
 *
 * Purpose: transform base64 encoded data to hex representation
 *
 * Arguments: pointer to the string which should be converted
 *
 * Returns: number of binary bytes or -1 if an error occurs
 *          the binary data is stored in the same buffer.
 *
 * Author: DG
 *
 */
int
base64tohex(char *buffer)
{
  int len;

  /* first decode base64, result is in buffer */
  len=base64decode(buffer);
  
  /* was there an error? */
  if (len==-1) return(-1);

  /* now transform binary to hex, result is again in buffer */
  return(bin2hex(buffer,len));
   
}
