/* $Id: rbt.c,v 1.2 2006/05/12 05:33:20 geschke Exp $
 * $Revision: 1.2 $ 
 *
** This file is part of the FLoP project but can be used separately
**
** Copyright (C) 2006 Dirk Geschke <Dirk@geschke-online.de>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* 
 * This is an implemention of a red-black binary tree
 * 
 * The functions and values for external programs are defined
 * in rbt.h. 
 *
 * Note: The function for removing a node is not implemented, yet.
 *
 * There are several things to do in order to use this RBtree:
 *
 *  1. adjust struct RBTvalue for your need, this is implemented
 *     as a linked list where further values to the same key can
 *     be stored. 
 *
 *  2. define type of key -> RBTkey in rbt.h
 *
 *  3. define a Tree in your application: 
 * 
 *      RBTree T; 
 *
 *     and initialize the root to be NULL:
 *
 *      T.root=NULL;
 *
 *  4. To find a key in the Tree you can invoke FindNode():
 * 
 *     int Result;
 *     RBnode * node;
 *
 *     node = FindNode(&T,&Result, key);
 *
 *     If the key is found then Result has the value 1 and the node
 *     points to the node containing the key: node->key==key.
 *
 *     Note: Due to the linked list of values you can use several
 *           values with the same key.
 *
 *     If the key is not found then Result has the value 0 and the
 *     node points to the new parent node
 *
 *  5. To insert a new key with the new value invoke InsertRBTkey 
 *     with the parent node (Note: value has to be allocated and
 *     initialized):
 *
 *     RBTvalue value;
 *     Result = InertRBTnode(&T,key,value,node)
 *
 *     The return value is either 0 if the new node could not be
 *     inserted. Otherwise the number of nodes in the tree is 
 *     returned.
 *
 *  6. To print the whole tree you can invoke PrintRBTree():
 *
 *     PrintRBTree(T.root);
 *
 */
#include <stdio.h>
#include <stdlib.h>

#define BLACK 0
#define RED 1

#include "rbt.h"

int LRotate(RBTree *T, RBTnode *x);
int LMoveUp(RBTree *T, RBTnode *y);
int BalanceNode(RBTree *T, RBTnode *x);
RBTnode * NewNode();

/*
 * Function: PrintRBTree(RBTnode *x)
 *
 * Purpose: Print the whole Tree recursive starting with
 *          node x. First the left tree is printed, then 
 *          the the right starting with the parent of the
 *          last left node.
 *
 *           level      value
 *             0          10
 *             1     5          15 
 *             2  3     7   12      17
 *
 *           would be printed as:
 *
 *           level: 0 value: 10 sum: 1 BLACK
 *           level: 1 value: 5 sum: 2  ....
 *           level: 2 value: 3 sum: 3  ....
 *           level: 2 value: 7 sum: 4  ....
 *           level: 1 value: 15 sum: 5 ....
 *           level: 2 value: 12 sum: 6 ....
 *           level: 2 value: 17 sum: 7 ....
 *
 *
 * Returns: 0 if all is ok....
 *
 * Author: DG
 */
int PrintRBTree(RBTnode *x)
{
  static int level=0;

#ifdef DEBUG
  printf("PrintRBTree called: %p\n", x);
#endif
  printf( "level: %i, value: %i ",level,x->key);
  if (x->color==BLACK ) {printf ("BLACK\n");}                     
  else {printf("RED\n");} 
  
  if (x->left!=NULL) {level++;PrintRBTree(x->left);level--;}
  if (x->right!=NULL){level++;PrintRBTree(x->right);level--;}
  return 0;
}


/*
 * Function: NewNode(RBTnode *node)
 *
 * Purpose: Allocate memory for a new node and preset some useful
 *          values. NewNodes are per default RED
 *
 * Returns: 1 on error -> malloc failed;
 *
 * Author: DG
 */

RBTnode * NewNode()
{
  RBTnode * node;
  /* allocate memory for node... */	
  node=(RBTnode *) malloc(sizeof(RBTnode));
  if (node == NULL) 
    {
      printf("malloc failed!\n");
      return NULL;
   }

#ifdef DEBUG
  printf("newnode at %p\n",node);
#endif

 /* ... and for RBTvalue */
 node->value=(RBTvalue *)malloc(sizeof(RBTvalue));

 if (node->value == NULL) 
   {
      printf ("malloc failed!\n");
      return NULL;
   }

 /* initialize to NULL, not necessary but helpful for validation checks */
  node->value->next=NULL;
  node->parent=NULL;
  node->left=NULL;
  node->right=NULL;

  node->key=0;
  /* default tree color is RED */
  node->color=RED;
  return node;
}

/*
 * Function: FindNode(T,x,key)
 *
 * Arguments: Tree T, pointer to node x, red-black key
 *
 * Purpose: find a node for given key 
 *
 * Returns: res contains 1 if the node is found, 
 *          then this node is returned
 *          0 if the node does not exist, in this
 *          case the parent node is returned
 *
 * Author: DG
 */
RBTnode * FindNode(RBTree *T, int *res, RBTkey key)
{
  RBTnode *y;
  RBTnode *x;

  y=T->root;
  if (T->root == NULL) 
    {
      /* this is an empty tree, we should not be
       * able to find the key...
       */
      *res=0; 
#ifdef DEBUG
      printf("Tree is empty!\n");
#endif
      return y;
    }
  while ( 1 )
    {
      x=y;
      if (y->key == key) 
	{
#ifdef debug
	  printf("y: %p. ykey %i , key %i\n",y,y->key,key);
#endif
	  *res=1;
	  return x;
	}
      if ( key > y->key)
	{
	  /* we must go right here */
	  y=y->right;
	}
      else
	{
	  y=y->left;
	}
      if (y == NULL) {*res=0; return x;}
    }
}


/*
 * Function: InsertRBTnode(T,key,value,parent)
 *
 * Arguments: Tree T, key, rb value 
 *
 * Purpose: inserts a new node to the Tree, the parent is returned
 *          by FindNode()
 *
 * Returns: 0 if the new node failed to insert.
 *          otherwise the number of entries in the cache
 *
 * Author: DG
 */
int InsertRBTnode(RBTree *T, RBTkey key,RBTvalue *value, RBTnode* parent)
{
  RBTnode *newnode;
  static int count=0;

#ifdef DEBUG
  printf("InsertRBTnode called with: %p, key: %i, value: %p, parent: %p\n",
	 T,key,value,parent);
#endif

  /* new entry */
  if (!(newnode=NewNode()))
    {
#ifdef DEBUG
      printf("failed to allocate new node\n");
#endif
      return 1;
    };

  newnode->key=key;
  /* node is parent, check where to insert */
  newnode->parent=parent;
  newnode->value=value;

  if (parent==NULL) 
    {
#ifdef DEBUG
      printf("Adding new root: %p\n",newnode);
#endif
      count=1;
      T->root=newnode;
      return count;
    }

  if (key > parent->key)
    {
      /* insert on the right */
      parent->right=newnode;
    }
  else
    {
      parent->left=newnode;
    }

  count++;

#ifdef DEBUG
  printf("new node at %p, parent: %p, left: %p, right %p\n", 
	 newnode,newnode->parent,newnode->left,newnode->right);
  printf("parent node at %p, parent: %p, left: %p, right %p\n", 
	 parent,parent->parent,parent->left,parent->right);
#endif

  /* Now Rebalance Tree */
  BalanceNode(T,newnode);
  
  return count;
}


/*
 * Function: RMoveUp(RBTree *T, RBTnode *y)
 *
 * Arguments: Tree T, node to be moved up y
 *
 * Purpose: Do a right move up of node y, this is sometimes
 *          called a right rotation:
 *
 *      z           z     |     z         z
 *       \           \    |    /         /
 *        x    ->     y   |   x    ->   y 
 *         \         /    |    \       /
 *          y       x     |     y     x
 *
 * Returns: 1 if move up fails, otherwise 0
 *
 * Author: DG
 */
int RMoveUp(RBTree *T,RBTnode *y)
{
   RBTnode *x;

   x=y->parent;

   if ((y==T->root) || (x->right==NULL))
     {
       printf("invalid RMoveUp: %p, %p, %p %i\n",y,T->root,y->left,y->key);
       return 1;
     }
   if (x==T->root) { T->root=y;}
   else
     {
       if (x==x->parent->left) {x->parent->left=y;}
       else {x->parent->right=y;}
     }
   y->parent=x->parent;
   x->parent=y;
   x->right=y->left;
   if (y->left) {y->left->parent=x;}
   y->left=x;
   return 0;
}


/*
 * Function: LMoveUp(RBTree *T, RBTnode *y)
 *
 * Arguments: Tree T, node to be moved up y
 *
 * Purpose: Do a left move up of node y, this is sometimes
 *          called a left rotation:
 *
 *     z         z       |       z         z
 *      \         \      |      /         /
 *       x    ->   y     |     x    ->   y 
 *      /            \   |    /           \   
 *     y              x  |   y             x
 *
 * Returns: 1 if move up fails, otherwise 0
 *
 * Author: DG
 */
int LMoveUp(RBTree *T,RBTnode *y)
{
  RBTnode *x;

  x=y->parent;

  if ((y==T->root) || (x->left==NULL))
     {
       printf("invalid LMoveUp: %p, %p, %p %i\n",y,T->root,y->right,y->key);
       return 1;
     }

   if (x==T->root) { T->root=y;}
   else
     {
       if (x==x->parent->left) {x->parent->left=y;}
       else {x->parent->right=y;}
     }
   y->parent=x->parent;
   x->parent=y;
   x->left=y->right;
   if (y->right) {y->right->parent=x;}
   y->right=x;
   return 0;
 }

/*
 * Function: LMoveUp(RBTree *T, RBTnode *x)
 *
 * Arguments: Tree T, new node x
 *
 * Purpose: Rebalance the red-black tree after insertion of node x
 *
 * Returns: 1 if fails (not yet possible), otherwise 0
 *
 * Author: DG
 */
int BalanceNode(RBTree *T, RBTnode *x)
{
  RBTnode *y;
  RBTnode *p;

  while ((x!=T->root) && (x->parent->color==RED))
    {
       y=x->parent; /* is RED! */

       /* stop condition, if parent is root we are done */
        if (y==T->root) 
          {
	    /* ensure root is black... */
      	    y->color=BLACK; 
            return 0; 
          } 

       p=y->parent; /* black */
       /*
        * x is RED and his parent is RED. If the uncle is RED, too, then
        * we can simply switch the colors, parent and uncle become BLACK 
        * and the Grandparent is recolored to RED. Then start the loop
        * again with x=y->parent (grandparent)!
        */
       if (p->left && p->right && p->left->color == RED && p->right->color==RED)
         {
           p->color=RED;
           p->left->color=BLACK;
           p->right->color=BLACK;
           x=p;
           continue;
         }

      if (y->parent->left==y)
        {
           if (y->right == x)
             {
                RMoveUp(T,x);
		LMoveUp(T,x);
		if (x->left) {x->left->color=BLACK;}
		x->color=RED;
		continue;
             }
	   else
	     LMoveUp(T,y);
         }
      else
        {
           if (y->left == x)
             {
                LMoveUp(T,x);
		RMoveUp(T,x);
		if (x->right) {x->right->color=BLACK;}
		x->color=RED;
		continue;
             }
	   else
	     RMoveUp(T,y);
         }

       /* now recolor childs to be black */
       if (y->left) {y->left->color=BLACK;}
       if (y->right) {y->right->color=BLACK;}
       y->color=RED;
       x=y;
     }
 /* to be sure */
 T->root->color=BLACK;
 return 0;
}

